#include "winver.h"

IDI_ICON1               ICON    DISCARDABLE     "icon.ico"

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 1,0,0,0
 PRODUCTVERSION 1,0,0,0
 FILEFLAGS 0x0L
 FILEFLAGSMASK 0x3fL
 FILEOS 0x00040004L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "000004b0"
        BEGIN
            VALUE "CompanyName", "The Qt Company Ltd"
            VALUE "FileDescription", "QtQuick Window Example"
            VALUE "FileVersion", "1.0.0.0"
            VALUE "LegalCopyright", "Copyright (C) 2015 The Qt Company Ltd."
            VALUE "InternalName", "window"
            VALUE "OriginalFilename", "window.exe"
            VALUE "ProductName", "QtQuick Window Example"
            VALUE "ProductVersion", "1.0.0.0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x0, 1200
    END
END
